import uuid
from datetime import datetime
from decimal import Decimal
from typing import Optional, TypedDict, Union

from billing.types._billing_entity import BillingEntity, BillingObject


class FeatureUsageSummary(BillingObject):
    feature_id: str
    feature_codename: str
    period_start: Optional[datetime]
    max_usage_limit: Optional[Decimal]
    used_amount: Optional[Decimal]


class FeatureUsageEvent(BillingEntity):
    feature_id: str
    user_id: uuid.UUID
    amount: Decimal
    refunded_at: Optional[datetime]


class FeatureRecordPayload(TypedDict):
    user_id: uuid.UUID
    amount: Union[Decimal, int]



import uuid
from typing import List, TypedDict

from typing_extensions import NotRequired

from billing.types._billing_entity import BillingEntity
from billing.types._customer import Customer


class Order(BillingEntity):
    pass


class OrderWithCustomer(Order):
    customer: Customer


class OrderCreatePayload(TypedDict):
    customer_id: str
    product_plan_ids: List[str]


class OrderListParams(TypedDict):
    user_id: NotRequired[uuid.UUID]



from decimal import Decimal
from typing import Optional

from billing.types._billing_entity import BillingEntityWithTimestamps
from billing.types._common import PeriodType, ProductPlanType
from billing.types._product import Product


class ProductPlan(BillingEntityWithTimestamps):
    product_id: str
    amount: Decimal
    currency: str
    period_type: Optional[PeriodType]
    period_count: int
    trial_period_days: Optional[int]
    product_plan_type: ProductPlanType


class ProductPlanWithProduct(ProductPlan):
    product: Product



from datetime import datetime
from typing import Any, ClassVar, Dict, Literal

from pydantic import VERSION as PYDANTIC_VERSION
from pydantic import BaseModel
from typing_extensions import Self

from billing._utils import json_loads

IS_PYDANTIC_V2 = PYDANTIC_VERSION.startswith("2.")

if IS_PYDANTIC_V2:
    from pydantic import ConfigDict


class BillingObject(BaseModel):
    """
    Data Transfer Object for a billing API object.
    """

    if IS_PYDANTIC_V2:
        model_config: ClassVar[ConfigDict] = ConfigDict(
            frozen=True,
        )
    else:

        class Config:
            frozen = True

    @classmethod
    def parse(cls, obj: Any) -> Self:
        if IS_PYDANTIC_V2:
            return cls.model_validate(obj)
        else:
            return cls.parse_obj(obj)

    def dump(self, mode: Literal["python", "json"] = "python") -> Dict[str, Any]:
        if IS_PYDANTIC_V2:
            return self.model_dump(mode=mode)
        elif mode == "python":
            return self.dict()
        else:
            return json_loads(self.json())  # type: ignore[no-any-return]


if IS_PYDANTIC_V2:
    from pydantic import RootModel

    class DynamicDictModel(BillingObject, RootModel[Dict[str, Any]]):

        @classmethod
        def init(cls, dict_: Dict[str, Any]) -> Self:
            return cls(dict_)

else:

    class DynamicDictModel(BillingObject):  # type: ignore[no-redef]
        __root__: Dict[str, Any]

        @classmethod
        def init(cls, dict_: Dict[str, Any]) -> Self:
            return cls(__root__=dict_)


class BillingEntity(BillingObject):
    id: str
    created_at: datetime


class BillingEntityWithTimestamps(BillingEntity):
    updated_at: datetime



from typing import List

from typing_extensions import Unpack

from billing.services._billing_api_service import BillingAPIService
from billing.types import Offer, OfferCancelPayload, OfferCustomerProductsBundle, OfferListParams


class OfferAPIService(BillingAPIService):
    def retrieve(self, object_id: str) -> OfferCustomerProductsBundle:
        return self._request(
            "GET",
            f"v1/offers/{object_id}/",
            data_model=OfferCustomerProductsBundle,
        )

    async def retrieve_async(self, object_id: str) -> OfferCustomerProductsBundle:
        return await self._request_async(
            "GET",
            f"v1/offers/{object_id}/",
            data_model=OfferCustomerProductsBundle,
        )

    def list(
        self,
        page_number: int = 1,
        page_size: int = 50,
        **params: Unpack[OfferListParams],
    ) -> List[Offer]:
        return self._request(
            "GET",
            "v1/offers/",
            params={
                "page": page_number,
                "page_size": page_size,
                **params,
            },
            data_model=Offer,
            batch_mode=True,
        )

    async def list_async(
        self,
        page_number: int = 1,
        page_size: int = 50,
        **params: Unpack[OfferListParams],
    ) -> List[Offer]:
        return await self._request_async(
            "GET",
            "v1/offers/",
            params={
                "page": page_number,
                "page_size": page_size,
                **params,
            },
            data_model=Offer,
            batch_mode=True,
        )

    def cancel(
        self,
        object_id: str,
        **payload: Unpack[OfferCancelPayload],
    ) -> None:
        return self._request(
            "POST",
            f"v1/offers/{object_id}/cancel/",
            json=payload,
        )

    async def cancel_async(
        self,
        object_id: str,
        **payload: Unpack[OfferCancelPayload],
    ) -> None:
        return await self._request_async(
            "POST",
            f"v1/offers/{object_id}/cancel/",
            json=payload,
        )



from billing.types._agreement import Agreement, AgreementListParams, AgreementWithTerms
from billing.types._billing_entity import BillingEntity, BillingEntityWithTimestamps, BillingObject
from billing.types._combined import (
    InvoiceWithItemsBundle,
    OfferCustomerProductsBundle,
    OfferInvoicesBundle,
    OfferPlansBundle,
    OrderOffersBundle,
    OrderWithOffersAndInvoicesBundle,
    ProductWithPlansAndImagesBundle,
)
from billing.types._common import HTTPMethod, PaymentMethod, PeriodType
from billing.types._customer import (
    Customer,
    CustomerAddress,
    CustomerAddressCreatePayload,
    CustomerCreatePayload,
    CustomerListParams,
)
from billing.types._feature import FeatureRecordPayload, FeatureUsageEvent, FeatureUsageSummary
from billing.types._image import Image
from billing.types._invoice import Invoice, InvoiceListParams
from billing.types._offer import Offer, OfferCancelPayload, OfferListParams, OfferWithCustomer
from billing.types._offer_product_plan import OfferProductPlan, OfferProductPlanWithProduct
from billing.types._order import Order, OrderCreatePayload, OrderListParams, OrderWithCustomer
from billing.types._product import Product, ProductDetailParams, ProductListParams
from billing.types._product_plan import ProductPlan, ProductPlanWithProduct
from billing.types._webhook_event import WebhookEvent

__all__ = (
    "Agreement",
    "AgreementListParams",
    "AgreementWithTerms",
    "BillingEntity",
    "BillingEntityWithTimestamps",
    "BillingObject",
    "InvoiceWithItemsBundle",
    "OfferPlansBundle",
    "OfferCustomerProductsBundle",
    "OfferInvoicesBundle",
    "OrderOffersBundle",
    "OrderWithOffersAndInvoicesBundle",
    "ProductWithPlansAndImagesBundle",
    "HTTPMethod",
    "PaymentMethod",
    "PeriodType",
    "Customer",
    "CustomerAddress",
    "CustomerAddressCreatePayload",
    "CustomerCreatePayload",
    "CustomerListParams",
    "FeatureUsageEvent",
    "FeatureRecordPayload",
    "FeatureUsageSummary",
    "Image",
    "Invoice",
    "InvoiceListParams",
    "Offer",
    "OfferCancelPayload",
    "OfferListParams",
    "OfferWithCustomer",
    "OfferProductPlan",
    "OfferProductPlanWithProduct",
    "Order",
    "OrderCreatePayload",
    "OrderListParams",
    "OrderWithCustomer",
    "Product",
    "ProductDetailParams",
    "ProductListParams",
    "ProductPlan",
    "ProductPlanWithProduct",
    "WebhookEvent",
)



import uuid
from datetime import datetime
from typing import Optional, TypedDict

from typing_extensions import NotRequired

from billing.types._billing_entity import BillingEntityWithTimestamps
from billing.types._common import CancellationFeedback
from billing.types._order import OrderWithCustomer


class Offer(BillingEntityWithTimestamps):
    order_id: str
    activated_at: Optional[datetime]
    cancel_at: Optional[datetime]
    cancellation_requested_at: Optional[datetime]
    cancellation_feedback: Optional[CancellationFeedback]
    cancellation_comment: Optional[str]


class OfferWithCustomer(Offer):
    order: OrderWithCustomer


class OfferCancelPayload(TypedDict):
    cancel_at: NotRequired[datetime]
    cancellation_feedback: NotRequired[CancellationFeedback]
    cancellation_comment: NotRequired[str]


class OfferListParams(TypedDict):
    user_id: NotRequired[uuid.UUID]
    order_id: NotRequired[str]
    offer_product_plan_id: NotRequired[str]



from typing import Optional


class BillingError(Exception):
    default_message: Optional[str] = None

    def __init__(self, message: Optional[str] = None) -> None:
        super().__init__(message or self.default_message)


class SignatureVerificationError(BillingError):
    pass


class BillingAPIError(BillingError):
    default_message = "API error"

    def __init__(
        self,
        message: Optional[str] = None,
        *,
        status_code: int,
        response_text: str,
    ) -> None:
        final_message = (
            f"Error message - {message or self.default_message}."
            f" Status code - {status_code}. Response text - {response_text}."
        )
        super().__init__(final_message)

        self._status_code = status_code
        self._response_text = response_text

    @property
    def status_code(self) -> int:
        return self._status_code

    @property
    def response_text(self) -> str:
        return self._response_text


class AuthenticationError(BillingAPIError):
    default_message = "Authentication error"


class RateLimitError(BillingAPIError):
    default_message = "Rate limit exceeded"


class FeatureUsageLimitExceededError(RateLimitError):
    default_message = "Feature usage limit exceeded"



import uuid
from datetime import datetime
from decimal import Decimal
from typing import Optional, TypedDict

from typing_extensions import NotRequired

from billing.types._billing_entity import BillingEntityWithTimestamps
from billing.types._common import PaymentMethod


class Invoice(BillingEntityWithTimestamps):
    total: Decimal
    currency: str
    issued_at: Optional[datetime]
    expired_at: datetime
    paid_at: Optional[datetime]
    refunded_at: Optional[datetime]
    payment_method: PaymentMethod
    receipt_url: str
    failed_reason: str
    checkout_session_url: Optional[str]


class InvoiceListParams(TypedDict):
    order_id: NotRequired[str]
    user_id: NotRequired[uuid.UUID]



from typing import List

from billing.types._invoice import Invoice
from billing.types._invoice_item import InvoiceItemWithProduct
from billing.types._offer import Offer, OfferWithCustomer
from billing.types._offer_product_plan import OfferProductPlan, OfferProductPlanWithProduct
from billing.types._order import Order
from billing.types._product import Product
from billing.types._product_plan import ProductPlan


class InvoiceWithItemsBundle(Invoice):
    items: List[InvoiceItemWithProduct]


class OfferPlansBundle(Offer):
    offer_product_plans: List[OfferProductPlan]


class OfferInvoicesBundle(Offer):
    invoices: List[InvoiceWithItemsBundle]


class OrderOffersBundle(Order):
    offers: List[OfferPlansBundle]


class OfferCustomerProductsBundle(OfferWithCustomer):
    last_invoice: InvoiceWithItemsBundle
    offer_product_plans: List[OfferProductPlanWithProduct]


class OrderWithOffersAndInvoicesBundle(Order):
    offers: List[OfferInvoicesBundle]


class ProductWithPlansAndImagesBundle(Product):
    product_plans: List[ProductPlan]



from typing import Any, Optional

from httpx import AsyncClient, Client

from billing._constants import DEFAULT_API_URL, TERMINAL_SECRET_KEY_PREFIX
from billing._http_client import HTTPClient
from billing.services import (
    AgreementAPIService,
    CustomerAPIService,
    FeatureAPIService,
    InvoiceAPIService,
    OfferAPIService,
    OrderAPIService,
    ProductAPIService,
)


class BillingClient:
    """
    Main class for interacting with the billing API.

    :param terminal_secret_key: Secret key for the terminal. Must starts with "tr_sk__".
    :param terminal_id: Unique identifier for the terminal.
    :param billing_api_url: URL of the billing API. Defaults to
        :const:`billing._constants.DEFAULT_API_URL`.
    :param max_network_retries: Maximum number of times to retry a request in case of a network error.
        Defaults to 0.
    :param setup_sync_client: Whether to set up a synchronous HTTP client. Defaults to False.
    :param setup_async_client: Whether to set up an asynchronous HTTP client. Defaults to True.
    :param sync_client: Existing synchronous HTTP client to use. Defaults to None.
    :param async_client: Existing asynchronous HTTP client to use. Defaults to None.
    :param httpx_client_kwargs: Additional keyword arguments to pass to the HTTPX client constructor.
    """

    def __init__(
        self,
        terminal_secret_key: str,
        terminal_id: str,
        *,
        billing_api_url: str = DEFAULT_API_URL,
        max_network_retries: int = 0,
        setup_sync_client: bool = False,
        setup_async_client: bool = True,
        sync_client: Optional[Client] = None,
        async_client: Optional[AsyncClient] = None,
        **httpx_client_kwargs: Any,
    ) -> None:
        if not terminal_secret_key.startswith(TERMINAL_SECRET_KEY_PREFIX):
            raise ValueError(f"Invalid terminal secret key. Must starts with `{TERMINAL_SECRET_KEY_PREFIX}`.")

        self._http_client = HTTPClient(
            terminal_secret_key=terminal_secret_key,
            terminal_id=terminal_id,
            base_url=billing_api_url,
            max_network_retries=max_network_retries,
            setup_sync_client=setup_sync_client,
            setup_async_client=setup_async_client,
            sync_client=sync_client,
            async_client=async_client,
            **httpx_client_kwargs,
        )

        self._agreements = AgreementAPIService(http_client=self._http_client)
        self._customers = CustomerAPIService(http_client=self._http_client)
        self._features = FeatureAPIService(http_client=self._http_client)
        self._invoices = InvoiceAPIService(http_client=self._http_client)
        self._offers = OfferAPIService(http_client=self._http_client)
        self._orders = OrderAPIService(http_client=self._http_client)
        self._products = ProductAPIService(http_client=self._http_client)

    @property
    def agreements(self) -> AgreementAPIService:
        return self._agreements

    @property
    def customers(self) -> CustomerAPIService:
        return self._customers

    @property
    def features(self) -> FeatureAPIService:
        return self._features

    @property
    def invoices(self) -> InvoiceAPIService:
        return self._invoices

    @property
    def offers(self) -> OfferAPIService:
        return self._offers

    @property
    def orders(self) -> OrderAPIService:
        return self._orders

    @property
    def products(self) -> ProductAPIService:
        return self._products

    def close_sync_client(self) -> None:
        self._http_client.close_sync_client()

    async def close_async_client(self) -> None:
        await self._http_client.close_async_client()



from billing.services._agreement_api_service import AgreementAPIService
from billing.services._billing_api_service import BillingAPIService
from billing.services._customer_api_service import CustomerAPIService
from billing.services._feature_api_service import FeatureAPIService
from billing.services._invoice_api_service import InvoiceAPIService
from billing.services._offer_api_service import OfferAPIService
from billing.services._order_api_service import OrderAPIService
from billing.services._product_service import ProductAPIService

__all__ = (
    "AgreementAPIService",
    "CustomerAPIService",
    "BillingAPIService",
    "FeatureAPIService",
    "InvoiceAPIService",
    "OfferAPIService",
    "OrderAPIService",
    "ProductAPIService",
)



DEFAULT_API_URL: str = "http://127.0.0.1:8000/api/"

TERMINAL_SECRET_KEY_PREFIX = "tr_sk__"
TERMINAL_WEBHOOK_SECRET_KEY_PREFIX = "tr_wsk__"

FEATURE_USAGE_LIMIT_EXCEEDED_ERROR_CODE = "FEATURE_USAGE_LIMIT_EXCEEDED_001"



from typing import TypedDict

from typing_extensions import NotRequired

from billing.types._billing_entity import BillingEntityWithTimestamps
from billing.types._common import ProductPlanType


class Product(BillingEntityWithTimestamps):
    name: str
    description: str


class ProductDetailParams(TypedDict):
    product_plan_currency: NotRequired[str]
    product_plan_type: NotRequired[ProductPlanType]


class ProductListParams(TypedDict):
    product_plan_currency: NotRequired[str]
    product_plan_type: NotRequired[ProductPlanType]



import uuid
from typing import Optional, TypedDict

from typing_extensions import NotRequired

from billing.types._billing_entity import BillingEntity, BillingObject


class CustomerAddress(BillingObject):
    country: str
    state: Optional[str] = None
    city: Optional[str] = None
    line1: Optional[str] = None
    line2: Optional[str] = None
    postal_code: Optional[str] = None


class Customer(BillingEntity):
    user_id: uuid.UUID
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    email: Optional[str] = None
    phone: Optional[str] = None
    address: Optional[CustomerAddress]


class CustomerAddressCreatePayload(TypedDict):
    country: str
    state: NotRequired[str]
    city: NotRequired[str]
    line1: NotRequired[str]
    line2: NotRequired[str]
    postal_code: NotRequired[str]


class CustomerCreatePayload(TypedDict):
    user_id: uuid.UUID
    first_name: NotRequired[str]
    last_name: NotRequired[str]
    email: NotRequired[str]
    phone: NotRequired[str]
    address: NotRequired[CustomerAddressCreatePayload]


class CustomerListParams(TypedDict):
    user_id: NotRequired[uuid.UUID]



from billing._billing_client import BillingClient
from billing._exceptions import (
    AuthenticationError,
    BillingAPIError,
    BillingError,
    FeatureUsageLimitExceededError,
    RateLimitError,
    SignatureVerificationError,
)
from billing._webhook import Webhook

__all__ = (
    "BillingClient",
    "AuthenticationError",
    "BillingAPIError",
    "BillingError",
    "FeatureUsageLimitExceededError",
    "RateLimitError",
    "SignatureVerificationError",
    "Webhook",
)
